export declare function query(field: string): Query;
export declare function select(array: any[]): Select;
type FilterFn = Parameters<typeof Array.prototype.filter>[0];
type SortFn = NonNullable<Parameters<typeof Array.prototype.sort>[0]>;
export declare class Query {
    protected _eval: string;
    protected _lookups: Map<number, any>;
    protected _sorts: Sort[];
    protected _limit: number;
    protected _offset: number;
    protected _expression: Expression | undefined;
    protected _filterCache: FilterFn | undefined;
    protected _sortCache: SortFn | undefined;
    protected _isEqual: typeof isEqual;
    constructor(field?: string);
    /**
     * Executes the query on the given array.
     */
    on<T = any>(input: PromiseLike<T[]>): PromiseLike<T[]>;
    on<T = any>(input: T[]): T[];
    /**
     * Executes the query on the given array and returns the first result.
     */
    first<T = any>(input: PromiseLike<T[]>): PromiseLike<T | undefined>;
    first<T = any>(input: T[]): T | undefined;
    /**
     * Executes the query on the given array and returns the last result.
     */
    last<T = any>(input: PromiseLike<T[]>): PromiseLike<T | undefined>;
    last<T = any>(input: T[]): T | undefined;
    /**
     * Ands two expressions or queries.
     */
    and(field?: string | Query): this;
    /**
     * Ors two expressions or queries.
     */
    or(field?: string | Query): this;
    /**
     * Checks if the value is not equal to the given value.
     */
    not(value?: any): this;
    /**
     * Checks if the value is equal to the given value.
     */
    equals(value: any): this;
    /**
     * Checks if the value is equal to the given value. Alias if `equals`.
     */
    is(value?: any): this;
    /**
     * Checks if the value is equal to the given value. Alias if `equals`.
     */
    isNaN(value?: any): this;
    /**
     * Checks if the value is not equal to the given value.
     */
    isnt(value: any): this;
    /**
     * Checks if the value is in the given array or Set.
     */
    within(value: any[] | Set<any>): this;
    /**
     * Checks if the array value has the given value.
     */
    has(value: any): this;
    /**
     * Checks if the array value has all the given values.
     */
    hasAll(value: any[]): this;
    /**
     * Checks if the value starts with the given value.
     */
    startsWith(value: string): this;
    /**
     * Checks if the value ends with the given value.
     */
    endsWith(value: string): this;
    /**
     * Checks if the value is greater than the given value.
     */
    gt(value: any): this;
    /**
     * Checks if the value is greater than or equal to the given value.
     */
    gte(value: any): this;
    /**
     * Checks if the value is less than the given value.
     */
    lt(value: any): this;
    /**
     * Checks if the value is less than or equal to the given value.
     */
    lte(value: any): this;
    /**
     * Tests if the value matches with the given regular expression.
     */
    test(value: RegExp): this;
    /**
     * Alias for `test`.
     */
    regex(value: RegExp): this;
    /**
     * Checks if the value is equivalent to the given value using `isEqual`.
     */
    same(value: any, options?: {
        partial?: boolean;
    }): this;
    /**
     * Filters by the given filter function.
     */
    filter(value: FilterFn): this;
    /**
     * Checks if any of the whole words are present in the given field.
     */
    search(words: string): this;
    /**
     * Checks if the value is of the given type or instance
     */
    type(value: Function | string): this;
    /**
     * Sorts the results by the given field.
     */
    sortOn(field: string): this;
    /**
     * Sorts the results by the given field. Alias of `sortOn` but `field` is not required.
     */
    sort(field?: string): this;
    /**
     * Sort the results in ascending order.
     */
    asc(): this | undefined;
    /**
     * Sort the results in descending order.
     */
    desc(): this | undefined;
    /**
     * Sort the results using the given function.
     */
    custom(value: SortFn): this | undefined;
    limit(value: number): this;
    offset(value: number): this;
    toString(): string;
    protected _getFilter(): ((value: any, index: number, array: any[]) => unknown) | null | undefined;
    protected _getSort(): ((a: any, b: any) => number) | null;
    protected _term(condition: string, field?: string | Query): this;
    protected _oper(operator: string, value?: any): this;
    protected _store(value: any): string;
    protected _sortDir(dir: number): this | undefined;
    protected _sortType(type: SortFn): this | undefined;
    protected _flush(): void;
}
export declare class Select extends Query {
    protected array: any[];
    constructor(array: any[]);
    where(field: string): void;
    end(): any[];
}
declare class Expression {
    term?: string | undefined;
    operator: string;
    value: any;
    not: boolean;
    template: string;
    constructor(term?: string | undefined);
    toString(): string;
}
declare class Sort {
    term?: string | undefined;
    type: SortFn;
    direction: number;
    constructor(term?: string | undefined);
    static regular(a: any, b: any): 0 | 1 | -1;
    toFunction(): (a: any, b: any) => number;
}
declare function isEqual(value: any, other: any, options: {
    partial?: boolean;
}): boolean;
export {};
